/*
 *   ScenarLang a domain specific language that attempts to make it even easier
 *   to write unit tests with mockito
 *    
 *   Copyright (C) 2012  X. Gillard 
 *   
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *   
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *   
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package xgi.ut.dsl.interpreters

import scala.collection.JavaConversions.seqAsJavaList
import scala.collection.mutable.Map

import xgi.ut.dsl.ObjectParsers.FieldRef
import xgi.ut.dsl.ObjectParsers.FnInvocation
import xgi.ut.dsl.ObjectParsers.MemberRef
import xgi.ut.dsl.ObjectParsers.Reference
import xgi.ut.dsl.ObjectParsers.SLiteral
import xgi.ut.dsl.ScenarParsers.Assignment
import xgi.ut.dsl.ScenarParsers.ConfigInstruction
import xgi.ut.dsl.ScenarParsers.Declaration
import xgi.ut.dsl.ScenarParsers.ExceptionInstruction
import xgi.ut.dsl.ScenarParsers.Instruction
import xgi.ut.dsl.ScenarParsers.SyntaxError
import xgi.ut.dsl.utl.ReflectionUtil.invoke
import xgi.ut.dsl.utl.ReflectionUtil.invokeException
import xgi.ut.dsl.utl.ReflectionWrapper.getFieldValue
import xgi.ut.dsl.utl.ReflectionWrapper.setField

/**
 * This trait provides the capability of interpreting a ScenarLang script.
 */
trait ScenarioInterpreter extends ToObject  {
  this:{
    val context:Any
    /* 
     * This is where we store the variables being defined in the scenario
     * This field needs to be declared outside the interpreter, that way we
     * can eventually plug-in some values declared at an other place.
     */
    val variables:Map[String, Any]
  } =>
  
  type ReflectionUtilConfigFn[ResultType, Outcome] = (Any, java.lang.String, java.util.List[_], ResultType) => Outcome
  
  /** interprets one instruction */
  def interpret(i:Instruction):Any = i match {
    case SyntaxError(msg) => Console.err.println(msg); throw new RuntimeException(msg)
    case ConfigInstruction(config) => addConfig(config)
    case Declaration(name, value) => {
      val valueObject = toObject(value)
      variables += name->valueObject
      valueObject
    }
    case Assignment(destination, value) => interpretComplexInstruction[Any, Object](invoke)(destination, value)
    case ExceptionInstruction(destination, value) => interpretComplexInstruction[Throwable, Any](invokeException)(destination, value)
  }
  
  /** interprets the results, isolates the actual target on which to apply the configFn and applies that function*/
  private[this] def interpretComplexInstruction[ResultType, Outcome](configFn:ReflectionUtilConfigFn[ResultType, Outcome])(destination:MemberRef, value:SLiteral):Any={
    val valueObject = toObject(value).asInstanceOf[ResultType]
    val source = dereference(destination.source, context)
    val interpretation = (source /: destination.invocations.dropRight(1))((newctx, elem) => dereference(elem, newctx))
    configureResult[ResultType, Outcome](configFn)(interpretation, destination.invocations.last, valueObject)
    valueObject
  }
  
  /** assigns a result to a given function invocation or field */
  private[this] def configureResult[ResultType, Outcome](configFn:ReflectionUtilConfigFn[ResultType, Outcome])(target:Any, ref:Reference, result:ResultType = null):Any = ref match {
    case MemberRef(src, destinations) => {
      val source = dereference(src, if(target!=null)target else context)
      val newTarget = (source /: destinations.dropRight(1))((newctx, elem) => dereference(elem, newctx))
      configureResult(configFn)(newTarget, destinations.last, result)
    }
    // an internal would make no sense here
    case FieldRef(name) => {
	  setField(target, name, result)
	  getFieldValue(target, name)
    }
    case FnInvocation(fname, params) => {
      val args = params map{toObject}
      configFn(target, fname, args, result)
    }
  }
  
}
